#!/usr/bin/env python3
import sys, os, gzip, argparse, textwrap

PROG = sys.argv[0].split('/')[-1]

class GeneAnnotation:
    def __init__(self, gene_id, transcript_id, protein_id):
        assert gene_id is not None
        self.id = gene_id
        self.transcript = transcript_id
        self.protein = protein_id
        if self.protein is None:
            self.protein = self.transcript
    def __str__(self):
        return f'{self.id}\t{self.transcript}\t{self.protein}'

#
# Command line options
#
def parse_args():
    p = argparse.ArgumentParser(prog=PROG)
    p.add_argument('-t', '--target-genes', required=True, help='Target Gene Table')
    p.add_argument('-o', '--out-dir', required=False, default='.', help='Output Directory')
    p.add_argument('-g', '--gtf', required=True, help='GTF Annotation')
    p.add_argument('-p', '--protein-fa', required=True, help='Protein FASTA')
    p.add_argument('-c', '--cds-fa', required=False, default=None, help='CDS FASTA')
    p.add_argument('-n', '--org-name', required=True, help='Name of the Organism')
    p.add_argument('-a', '--assembly-id', required=True, help='Assembly ID (used to name outputs)')
    # Check input arguments
    args = p.parse_args()
    args.out_dir = args.out_dir.rstrip('/')
    if not os.path.exists(args.out_dir):
        sys.exit(f"Error: '{args.out_dir}' does not exist.")
    if not os.path.exists(args.gtf):
        sys.exit(f"Error: '{args.gtf}' does not exist.")
    if not os.path.exists(args.protein_fa):
        sys.exit(f"Error: '{args.protein_fa}' does not exist.")
    if args.cds_fa is not None:
        if not os.path.exists(args.cds_fa):
            sys.exit(f"Error: '{args.cds_fa}' does not exist.")
    if not os.path.exists(args.target_genes):
        sys.exit(f"Error: '{args.target_genes}' does not exist.")
    return args

def parse_ncbi_cds_fasta_id(seq_id):
    # print(seq_id)
    # The IDs look like this:  >lcl|NC_047503.1_cds_XP_033950052.1_10
    new_seq_id = seq_id.split('|')[1].split('_')
    if len(new_seq_id) == 4:
        new_seq_id = new_seq_id[2]
    elif len(new_seq_id) == 6:
        new_seq_id = new_seq_id[3:5]
        new_seq_id = '_'.join(new_seq_id)
    else:
        new_seq_id = None
    # print(new_seq_id)
    return new_seq_id

def parse_target_gene_tbl(target_gene_tbl):
    target_genes = dict()
    with open(target_gene_tbl) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            fields = line.strip('\n').split('\t')
            gene_id = fields[0]
            gene_name = fields[1]
            target_genes[gene_id] = gene_name
    print(f'Processing {len(target_genes):,} target genes.')
    return target_genes

def extract_gtf_attributes(attributes, attribute_id='gene_id'):
    selected_attributes = ('gene_id', 'transcript_id', 'protein_id')
    assert attribute_id in selected_attributes
    for attribute in attributes:
        attribute = attribute.lstrip(' ').rstrip(' ')
        if len(attribute) == 0:
            continue
        attribute_type = attribute.split(' ')[0]
        if attribute_type != attribute_id:
            continue
        attribute_value = attribute.split(' ')[1].strip('"')
        if len(attribute_value) == 0:
            return None
        return attribute_value

def parse_gtf_file(gtf_f, target_genes, assembly_id, outdir='.'):
    # Create output GTF
    outf = open(f'{outdir}/{assembly_id}.gtf', 'w')
    gtf_genes = dict()
    target_gene_set = set(target_genes.keys())
    fh = None
    if gtf_f.endswith('gz'):
        fh = gzip.open(gtf_f, 'rt')
    else:
        fh = open(gtf_f)
    with fh:
        previous_gene = None
        gene_id = None
        transcript_id = None
        protein_id = None
        rows = 0
        for i, line in enumerate(fh):
            if line.startswith('#'):
                continue
            fields = line.strip('\n').split('\t')
            feature = fields[2]
            attributes = fields[8].split(';')
            source = fields[1]
            # Skipping unwanted features
            if feature not in ('CDS', 'exon', 'gene', 'start_codon', 'stop_codon', 'transcript'):
                continue
            # Extract gene id of all rows to compare against target
            gene_id = extract_gtf_attributes(attributes, 'gene_id')
            if gene_id is None:
                continue
            if gene_id in target_gene_set:
                outf.write(line)
                rows += 1
                # Process info when reaching a new gene in the file
                if gene_id != previous_gene:
                    if previous_gene is not None:
                        gene_annotation = GeneAnnotation(previous_gene, transcript_id, protein_id)
                        gtf_genes.setdefault(gene_annotation.id, list())
                        gtf_genes[gene_annotation.id].append(gene_annotation)
                    previous_gene = gene_id
                # When working on a target gene, extract other features
                if feature == 'CDS':
                    transcript_id = extract_gtf_attributes(attributes, 'transcript_id')
                    protein_id = extract_gtf_attributes(attributes, 'protein_id')
                    # To handle the NCBI CDS fasta seqs
                    if transcript_id.startswith('XM') and source in ['Gnomon', 'Genbank']:
                        transcript_id = protein_id
                    elif transcript_id.startswith('gnl') and source in ['Gnomon', 'Genbank']:
                        transcript_id = protein_id
        # Add the last gene
        gene_annotation = GeneAnnotation(previous_gene, transcript_id, protein_id)
        gtf_genes.setdefault(gene_annotation.id, list())
        gtf_genes[gene_annotation.id].append(gene_annotation)
    assert set(gtf_genes.keys()) == target_gene_set
    print(f'Saved {len(gtf_genes):,} total genes from the input GTF file ({rows:,} lines total).')
    outf.close()
    return gtf_genes

def wrap_sequence(sequence_str: str, columns: int = 80):
    assert columns > 50
    wrapped_list = textwrap.wrap(sequence_str, columns)
    new_seq = '\n'.join(wrapped_list)
    return new_seq

def process_fasta(fasta_f: str, target_set: set):
    # Extracted sequences
    extracted_sequeces = dict()
    # Open input
    fh = None
    if fasta_f.endswith('gz'):
        fh = gzip.open(fasta_f, 'rt')
    else:
        fh = open(fasta_f, 'r')
    with fh:
        previous_seq = None
        seq_id = None
        sequence = ''
        for i, line in enumerate(fh):
            if line.startswith('#'):
                continue
            line = line.strip('\n')
            if line.startswith('>'):
                seq_id = line.lstrip('>').split(' ')[0]
                # To process the NCBI CDS fasta seqs
                if seq_id.startswith('lcl'):
                    seq_id = parse_ncbi_cds_fasta_id(seq_id)
                if previous_seq in target_set:
                    extracted_sequeces[previous_seq] = sequence
                else:
                    try:
                        new_prev_seq = previous_seq.split('.')[0]
                        if new_prev_seq in target_set:
                            extracted_sequeces[new_prev_seq] = sequence
                    except:
                        pass
                previous_seq = seq_id
                sequence = ''
            else:
                sequence += line
        if previous_seq in target_set:
            extracted_sequeces[previous_seq] = sequence
        else:
            try:
                new_prev_seq = previous_seq.split('.')[0]
                if new_prev_seq in target_set:
                    extracted_sequeces[new_prev_seq] = sequence
            except:
                pass
    return extracted_sequeces

def extract_peptide_fasta(fasta_f, target_genes, gtf_genes, assembly_id, org_id, outdir='.'):
    # Prepare the set of input proteins
    proteins = dict()
    for gene_id in gtf_genes:
        gene = gtf_genes[gene_id]
        for annots in gene:
            assert isinstance(annots, GeneAnnotation)
            prot = annots.protein
            proteins[prot] = gene_id
    # Extract the target sequences
    extracted_proteins = process_fasta(fasta_f, set(proteins.keys()))
    # Create output FASTA
    outf = open(f'{outdir}/{assembly_id}.peptide.fa', 'w')
    for prot_id in extracted_proteins:
        sequence = extracted_proteins[prot_id]
        sequence = wrap_sequence(sequence, 80)
        gene_id = proteins[prot_id]
        gene_name = target_genes[gene_id]
        fasta = f'>{prot_id} gene:"{gene_id}" name:"{gene_name}" org:"{org_id}"\n{sequence}\n'
        outf.write(fasta)
    outf.close()
    print(f'Extracted {len(extracted_proteins):,} sequences from the peptide FASTA.')

def extract_cds_fasta(fasta_f, target_genes, gtf_genes, assembly_id, org_id, outdir='.'):
    # Prepare the set of input CDSs
    cds = dict()
    for gene_id in gtf_genes:
        gene = gtf_genes[gene_id]
        for annots in gene:
            assert isinstance(annots, GeneAnnotation)
            transcript = annots.transcript
            cds[transcript] = gene_id
    # Extract the target sequences
    extracted_cds = process_fasta(fasta_f, set(cds.keys()))
    # Create output FASTA
    outf = open(f'{outdir}/{assembly_id}.cds.fa', 'w')
    for cds_id in extracted_cds:
        sequence = extracted_cds[cds_id]
        sequence = wrap_sequence(sequence, 80)
        gene_id = cds[cds_id]
        gene_name = target_genes[gene_id]
        fasta = f'>{cds_id} gene:"{gene_id}" name:"{gene_name}" org:"{org_id}"\n{sequence}\n'
        outf.write(fasta)
    outf.close()
    print(f'Extracted {len(extracted_cds):,} sequences from the CDS FASTA.')

def main():
    argv_str = ' '.join(sys.argv)
    print(f'{argv_str}\n')
    args = parse_args()
    target_genes = parse_target_gene_tbl(args.target_genes)
    gtf_genes = parse_gtf_file(args.gtf, target_genes, args.assembly_id, args.out_dir)
    extract_peptide_fasta(args.protein_fa, target_genes, gtf_genes, args.assembly_id, args.org_name, args.out_dir)
    if args.cds_fa is not None:
        extract_cds_fasta(args.cds_fa, target_genes, gtf_genes, args.assembly_id, args.org_name, args.out_dir)

# Run Code
if __name__ == '__main__':
    main()
